@@ -1,5 +1,8 @@
 from __future__ import annotations  # allow forward references
 import sys
+from concurrent import futures
+from concurrent.futures import Future, ThreadPoolExecutor
+from enum import StrEnum
 from importlib import import_module
 from inspect import FrameInfo, stack
 from enum import Enum
@@ -11,12 +14,12 @@ from pypomes_db import (
     db_select, db_insert, db_update, db_delete
 )
 from types import ModuleType
-from typing import Any, Type, TypeVar
+from typing import Any, TypeVar
 
 from .sob_config import (
-    SOB_BASE_FOLDER,
-    sob_db_specs, sob_attrs_map,
-    sob_attrs_unique, sob_cls_references
+    SOB_BASE_FOLDER, SOB_MAX_THREADS,
+    sob_db_specs, sob_col_names,
+    sob_attrs_map, sob_attrs_unique
 )
 
 # 'Sob' stands for all subclasses of 'PySob'
@@ -30,11 +33,12 @@ class PySob:
     """
 
     def __init__(self,
-                 errors: list[str] = None,
-                 load_references: bool = False,
+                 __references: list[type] = None,
+                 /,
                  where_data: dict[str, Any] = None,
                  db_engine: DbEngine = None,
                  db_conn: Any = None,
+                 errors: list[str] = None,
                  logger: Logger = None) -> None:
 
         self._logger: Logger = logger
@@ -44,131 +48,149 @@ class PySob:
         # determine whether this instance exists in the database
         self._is_new: bool = True
 
+        # make sure to have an errors list
+        if not isinstance(errors, list):
+            errors = []
+
         if where_data:
             self.set(data=where_data)
-            self.load(errors=errors,
+            self.load(__references,
                       omit_nulls=True,
                       db_engine=db_engine,
-                      db_conn=db_conn)
-        if not errors and load_references:
-            self.load_references(errors=errors,
-                                 db_engine=db_engine,
-                                 db_conn=db_conn)
+                      db_conn=db_conn,
+                      errors=errors)
 
     def insert(self,
-               errors: list[str] = None,
                db_engine: DbEngine = None,
-               db_conn: Any = None) -> bool:
+               db_conn: Any = None,
+               committable: bool = None,
+               errors: list[str] = None) -> bool:
 
-        # prepara data for INSERT
+        # prepare data for INSERT
         return_col: dict[str, type] | None = None
         insert_data: dict[str, Any] = self.to_columns(omit_nulls=True)
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
-        if sob_db_specs[class_name][3]:
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        is_identity: bool = sob_db_specs[cls_name][2]
+        if is_identity:
             # PK is an identity column
-            insert_data.pop(sob_db_specs[class_name][1], None)
-            return_col = {sob_db_specs[class_name][1]: sob_db_specs[class_name][2]}
+            pk_name: str = sob_col_names[cls_name][0]
+            pk_type: type = sob_db_specs[cls_name][1]
+            insert_data.pop(pk_name, None)
+            return_col = {pk_name: pk_type}
 
         # make sure to have an errors list
         if not isinstance(errors, list):
             errors = []
 
         # execute the INSERT statement
-        rec: tuple[Any] = db_insert(errors=errors,
-                                    insert_stmt=f"INSERT INTO {sob_db_specs[class_name][0]}",
+        tbl_name = sob_db_specs[cls_name][0]
+        rec: tuple[Any] = db_insert(insert_stmt=f"INSERT INTO {tbl_name}",
                                     insert_data=insert_data,
                                     return_cols=return_col,
                                     engine=db_engine,
                                     connection=db_conn,
+                                    committable=committable,
+                                    errors=errors,
                                     logger=self._logger)
         if not errors:
             self._is_new = False
-            if sob_db_specs[class_name][3]:
+            if is_identity:
                 # PK is an identity column
                 self.id = rec[0]
         elif self._logger:
             self._logger.error(msg="Error INSERTing into table "
-                                   f"{sob_db_specs[class_name][0]}: {'; '.join(errors)}")
+                                   f"{tbl_name}: {'; '.join(errors)}")
         return not errors
 
     def update(self,
-               errors: list[str] = None,
                db_engine: DbEngine = None,
-               db_conn: Any = None) -> bool:
+               db_conn: Any = None,
+               committable: bool = None,
+               errors: list[str] = None) -> bool:
 
         # prepare data for UPDATE
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
         update_data: dict[str, Any] = self.to_columns(omit_nulls=False)
-        key: int = update_data.pop(sob_db_specs[class_name][1])
+        tbl_name: str = sob_db_specs[cls_name][0]
+        pk_name = sob_col_names[cls_name][0]
+        key: int | str = update_data.pop(pk_name)
 
         # make sure to have an errors list
         if not isinstance(errors, list):
             errors = []
 
         # execute the UPDATE statement
-        db_update(errors=errors,
-                  update_stmt=f"UPDATE {sob_db_specs[class_name][0]}",
+        db_update(update_stmt=f"UPDATE {tbl_name}",
                   update_data=update_data,
-                  where_data={sob_db_specs[class_name][1]: key},
+                  where_data={pk_name: key},
                   min_count=1,
                   max_count=1,
                   engine=db_engine,
                   connection=db_conn,
+                  committable=committable,
+                  errors=errors,
                   logger=self._logger)
 
         if errors and self._logger:
             self._logger.error(msg="Error UPDATEing table "
-                                   f"{sob_db_specs[class_name][0]}: {'; '.join(errors)}")
+                                   f"{tbl_name}: {'; '.join(errors)}")
         return not errors
 
     def persist(self,
-                errors: list[str] = None,
                 db_engine: DbEngine = None,
-                db_conn: Any = None) -> bool:
+                db_conn: Any = None,
+                committable: bool = None,
+                errors: list[str] = None) -> bool:
 
         # declare the return variable
         result: bool
 
         if self._is_new:
-            result = self.insert(errors=errors,
-                                 db_engine=db_engine,
-                                 db_conn=db_conn)
+            result = self.insert(db_engine=db_engine,
+                                 db_conn=db_conn,
+                                 committable=committable,
+                                 errors=errors)
         else:
-            result = self.update(errors=errors,
-                                 db_engine=db_engine,
-                                 db_conn=db_conn)
+            result = self.update(db_engine=db_engine,
+                                 db_conn=db_conn,
+                                 committable=committable,
+                                 errors=errors)
         return result
 
     def delete(self,
-               errors: list[str] = None,
                db_engine: DbEngine = None,
-               db_conn: Any = None) -> int | None:
+               db_conn: Any = None,
+               committable: bool = None,
+               errors: list[str] = None) -> int | None:
 
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
         where_data: dict[str, Any]
+        tbl_name: str = sob_db_specs[cls_name][0]
+        pk_name: str = sob_col_names[cls_name][0]
         if self.id:
-            where_data = {sob_db_specs[class_name][1]: self.id}
+            where_data = {pk_name: self.id}
         else:
             where_data = self.to_columns(omit_nulls=True)
-            where_data.pop(sob_db_specs[class_name][1], None)
+            where_data.pop(pk_name, None)
 
         # make sure to have an errors list
         if not isinstance(errors, list):
             errors = []
 
         # execute the DELETE statement
-        result: int = db_delete(errors=errors,
-                                delete_stmt=f"DELETE FROM {sob_db_specs[class_name][0]}",
+        result: int = db_delete(delete_stmt=f"DELETE FROM {tbl_name}",
                                 where_data=where_data,
                                 max_count=1,
                                 engine=db_engine,
                                 connection=db_conn,
+                                committable=committable,
+                                errors=errors,
                                 logger=self._logger)
         if not errors:
             self.clear()
         elif self._logger:
             self._logger.error(msg="Error DELETEing from table "
-                                   f"{sob_db_specs[class_name][0]}: {'; '.join(errors)}")
+                                   f"{tbl_name}: {'; '.join(errors)}")
         return result
 
     def clear(self) -> None:
@@ -179,9 +201,10 @@ class PySob:
     def set(self,
             data: dict[str, Any]) -> None:
 
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        tbl_name = sob_db_specs[cls_name][0]
         for key, value in data.items():
-            attr: str = (sob_attrs_map.get(class_name) or {}).get(key) or key
+            attr: str = (sob_attrs_map.get(cls_name) or {}).get(key) or key
 
             # use enum names assigned as values in 'data'
             if isinstance(value, Enum) and "use_names" in value.__class__:
@@ -190,68 +213,73 @@ class PySob:
             if attr in self.__dict__:
                 self.__dict__[attr] = value
             elif self._logger:
-                self._logger.warning(msg=f"'{attr}'is not an attribute of "
-                                         f"{sob_db_specs[class_name][0]}")
+                self._logger.warning(msg=f"'{attr}'is not an attribute of {tbl_name}")
 
     def is_new(self) -> bool:
 
         return self._is_new
 
     def is_in_db(self,
-                 errors: list[str] = None,
                  db_engine: DbEngine = None,
-                 db_conn: Any = None) -> bool | None:
+                 db_conn: Any = None,
+                 errors: list[str] = None) -> bool | None:
 
         # initialize the return variable
         result: bool = False
 
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        tbl_name: str = sob_db_specs[cls_name][0]
+        pk_name: str = sob_col_names[cls_name][0]
         where_data: dict[str, Any] | None = None
         if self.id:
             # use object's ID
-            where_data = {sob_db_specs[class_name][1]: self.id}
-        elif sob_attrs_unique[class_name]:
+            where_data = {pk_name: self.id}
+        elif sob_attrs_unique[cls_name]:
             # use first set of unique attributes with non-null values found
-            for attr_set in sob_attrs_unique[class_name]:
+            for attr_set in sob_attrs_unique[cls_name]:
                 attrs_unique: dict[str, Any] = {}
                 for attr in attr_set:
                     val: Any = self.__dict__.get(attr)
                     if val is not None:
                         attrs_unique[attr] = val
-                if len(attrs_unique) == len(sob_attrs_unique[class_name]):
+                if len(attrs_unique) == len(sob_attrs_unique[cls_name]):
                     where_data = attrs_unique
                     break
 
         if not where_data:
             # use object's available data
             where_data = self.to_columns(omit_nulls=True)
-            where_data.pop(sob_db_specs[class_name][1], None)
+            where_data.pop(pk_name, None)
 
         if not where_data:
-            result = db_exists(errors=errors,
-                               table=sob_db_specs[class_name][0],
+            result = db_exists(table=tbl_name,
                                where_data=where_data,
                                engine=db_engine,
                                connection=db_conn,
+                               errors=errors,
                                logger=self._logger)
         return result
 
     def load(self,
-             errors: list[str] = None,
+             __references: list[type] = None,
+             /,
              omit_nulls: bool = True,
              db_engine: DbEngine = None,
-             db_conn: Any = None) -> bool:
+             db_conn: Any = None,
+             errors: list[str] = None) -> bool:
 
         # initialize the return variable
         result: bool = False
 
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        tbl_name: str = sob_db_specs[cls_name][0]
+        pk_name: str = sob_col_names[cls_name][0]
         where_data: dict[str, Any]
         if self.id:
-            where_data = {sob_db_specs[class_name][1]: self.id}
+            where_data = {pk_name: self.id}
         else:
             where_data = self.to_columns(omit_nulls=omit_nulls)
-            where_data.pop(sob_db_specs[class_name][1], None)
+            where_data.pop(pk_name, None)
 
         # make sure to have an errors list
         if not isinstance(errors, list):
@@ -259,24 +287,24 @@ class PySob:
 
         # loading the object from the database might fail
         attrs: list[str] = self.get_columns()
-        recs: list[tuple] = db_select(errors=errors,
-                                      sel_stmt=f"SELECT {', '.join(attrs)} "
-                                               f"FROM {sob_db_specs[class_name][0]}",
+        recs: list[tuple] = db_select(sel_stmt=f"SELECT {', '.join(attrs)} "
+                                               f"FROM {tbl_name}",
                                       where_data=where_data,
                                       limit_count=2,
                                       engine=db_engine,
                                       connection=db_conn,
+                                      errors=errors,
                                       logger=self._logger)
         msg: str | None = None
         if errors:
             msg = ("Error SELECTing from table "
-                   f"{sob_db_specs[class_name][0]}: {'; '.join(errors)}")
+                   f"{tbl_name}: {'; '.join(errors)}")
         elif not recs:
             msg = (f"No record found on table "
-                   f"{sob_db_specs[class_name][0]} for {dict_stringify(where_data)}")
+                   f"{tbl_name} for {dict_stringify(where_data)}")
         elif len(recs) > 1:
             msg = (f"More than on record found on table "
-                   f"{sob_db_specs[class_name][0]} for {dict_stringify(where_data)}")
+                   f"{tbl_name} for {dict_stringify(where_data)}")
 
         if msg:
             errors.append(msg)
@@ -286,30 +314,33 @@ class PySob:
             rec: tuple = recs[0]
             for inx, attr in enumerate(attrs):
                 # PK attribute in DB table might have a different name
-                if attr == sob_db_specs[class_name][0]:
+                if attr == pk_name:
                     self.__dict__["id"] = rec[inx]
                 else:
                     self.__dict__[attr] = rec[inx]
             self._is_new = False
-            result = True
+            if __references:
+                self.load_references(__references,
+                                     db_engine=db_engine,
+                                     db_conn=db_conn,
+                                     errors=errors)
+            result = not errors
 
         return result
 
     def get_columns(self) -> list[str]:
 
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
-        # PK attribute in DB table might have a different name
-        result: list[str] = [sob_db_specs[class_name][1]]
-        result.extend([k for k in self.__dict__
-                      if k.islower() and not k.startswith("_") and not k == "id"])
-        return result
+        # PK attribute in DB table might have a different : list[str] = []
+        return [k for k in self.__dict__
+                if k.islower() and not k.startswith("_")]
 
     def to_columns(self,
                    omit_nulls: bool) -> dict[str, Any]:
 
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        pk_name: str = sob_col_names[cls_name][0]
         # PK attribute in DB table might have a different name
-        result: dict[str, Any] = {sob_db_specs[class_name][1]: self.__dict__.get("id")}
+        result: dict[str, Any] = {pk_name: self.__dict__.get("id")}
         result.update({k: v for k, v in self.__dict__.items()
                       if k.islower() and not (k.startswith("_") or k == "id" or (omit_nulls and v is None))})
         return result
@@ -327,70 +358,76 @@ class PySob:
         # initialize the return variable
         result: dict[str, Any] = {}
         
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
         for k, v in data.items():
             if not omit_nulls or v is not None:
-                attr: str = dict_get_key(source=sob_attrs_map.get(class_name) or {},
+                attr: str = dict_get_key(source=sob_attrs_map.get(cls_name) or {},
                                          value=k) or k
                 result[attr] = v
 
         return result
 
-    # noinspection PyUnusedLocal
     def load_reference(self,
-                       __cls: type[Sob],
+                       __reference: type,
                        /,
-                       errors: list[str] = None,
                        db_engine: DbEngine = None,
-                       db_conn: Any = None) -> None:
+                       db_conn: Any = None,
+                       errors: list[str] = None) -> None:
+
+        self.load_references([__reference],
+                             db_engine=db_engine,
+                             db_conn=db_conn,
+                             errors=errors)
+
+    # noinspection PyUnusedLocal
+    def load_references(self,
+                        __references: list[type],
+                        /,
+                        db_engine: DbEngine = None,
+                        db_conn: Any = None,
+                        errors: list[str] = None) -> None:
 
         # must be implemented by subclasses containing references
-        msg: str = f"Subclass {__cls.__module__}.{__cls.__qualname__} failed to implement 'load_reference()'"
+        msg: str = (f"Subclass {self.__class__.__module__}.{self.__class__.__qualname__} "
+                    "failed to implement 'load_references()'")
         if isinstance(errors, list):
             errors.append(msg)
         if self._logger:
             self._logger.error(msg=msg)
 
-    def load_references(self,
-                        errors: list[str] = None,
-                        db_engine: DbEngine = None,
-                        db_conn: Any = None) -> None:
+    def invalidate_reference(self,
+                             __reference: type,
+                             /,
+                             db_engine: DbEngine = None,
+                             db_conn: Any = None,
+                             errors: list[str] = None) -> None:
 
-        # make sure to have an errors list
-        if not isinstance(errors, list):
-            errors = []
+        self.invalidate_references([__reference],
+                                   db_engine=db_engine,
+                                   db_conn=db_conn,
+                                   errors=errors)
 
-        # traverse the references
-        class_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
-        for name in (sob_cls_references.get(class_name) or []):
-            # identify the reference
-            pos: int = name.rfind(".")
-            module_name: str = name[:pos]
-            module: ModuleType = import_module(name=module_name)
-            cls: type[Sob] = getattr(module,
-                                     name[pos+1:])
-            # load it
-            self.load_reference(cls,
-                                errors=errors,
-                                db_engine=db_engine,
-                                db_conn=db_conn)
+    # noinspection PyUnusedLocal
+    def invalidate_references(self,
+                              __references: list[type],
+                              /,
+                              db_engine: DbEngine = None,
+                              db_conn: Any = None,
+                              errors: list[str] = None) -> None:
 
-            if errors and self._logger:
-                self._logger.error(msg=f"Error SELECTing from table "
-                                       f"{sob_db_specs[name][0]}: {'; '.join(errors)}")
-                break
+        # must be implemented by subclasses containing references
+        msg: str = (f"Subclass {self.__class__.__module__}.{self.__class__.__qualname__} "
+                    "failed to implement 'invalidate_references()'")
+        if isinstance(errors, list):
+            errors.append(msg)
+        if self._logger:
+            self._logger.error(msg=msg)
 
     @staticmethod
-    # HAZARD:
-    #   1. because 'typings.Type' has been deprecated, 'type' should be used here
-    #   2. 'Sob' stands for all subclasses of 'PySob', and thus 'type[Sob]' should suffice
-    #   3. PyCharm's code inspector, however, takes 'type[Sob]' to mean strict 'PySob' class
-    #   4. thus, a fallback to 'Type[PySub]' was necessary
-    def initialize(db_specs: tuple[str, str, int | str] |
-                             tuple[str,  str, int, bool],  # noqa
+    def initialize(db_specs: tuple[type[StrEnum] | list[str], int | str] |
+                             tuple[type[StrEnum] | list[str], int, bool],  # noqa
                    attrs_map: dict[str, str] = None,
                    attrs_unique: list[tuple[str]] = None,
-                   sob_references: list[Type[PySob]] = None,
                    logger: Logger = None) -> None:
 
         # obtain the invoking class
@@ -399,28 +436,42 @@ class PySob:
                                                     logger=logger)
         # initialize its data
         if cls:
-            name: str = f"{cls.__module__}.{cls.__qualname__}"
-            if len(db_specs) == 3:
-                # 'id' defaults to being an identity attribute in the DB for type 'int'
-                db_specs += (db_specs[2] is int,)
-            sob_db_specs.update({name: db_specs})
+            # retrieve the list of DB names
+            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+            attrs: list[str] = [attr.value for attr in db_specs[0]] \
+                if isinstance(db_specs[0], type) else db_specs[0].copy()
+            tbl_name: str = attrs.pop(0)
+
+            # register the names of DB columnns
+            sob_col_names.update({cls_name: tuple(attrs)})
+
+            # register the DB specs (table, PK type, PK entity state)
+            if len(db_specs) == 2:
+                # PK defaults to being an identity attribute in the DB for type 'int'
+                db_specs += (db_specs[1] is int,)
+            sob_db_specs.update({cls_name: (tbl_name, db_specs[1], db_specs[2])})
+
+            # register the mapping of input parameters to DB columns
             if attrs_map:
-                sob_attrs_map.update({name: attrs_map})
+                sob_attrs_map.update({cls_name: attrs_map})
+
+            # register the sets of unique attributes
             if attrs_unique:
-                sob_attrs_unique.update({name: attrs_unique})
-            if sob_references:
-                ref_names: list[str] = [f"{c.__module__}.{c.__qualname__}" for c in sob_references]
-                sob_cls_references.update({name: ref_names})
+                sob_attrs_unique.update({cls_name: attrs_unique})
+
             if logger:
-                logger.debug(msg=f"Inicialized access data for class '{name}'")
+                logger.debug(msg=f"Inicialized access data for class '{cls_name}'")
 
     @staticmethod
-    def count(errors: list[str] = None,
+    def count(table_alias: str = None,
+              table_joins: str = None,
+              count_clause: str = None,
               where_clause: str = None,
               where_vals: tuple = None,
               where_data: dict[str, Any] = None,
               db_engine: DbEngine = None,
               db_conn: Any = None,
+              errors: list[str] = None,
               logger: Logger = None) -> int | None:
 
         # inicialize the return variable
@@ -434,14 +485,20 @@ class PySob:
         cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
                                                     logger=logger)
         if not errors:
-            name: str = f"{cls.__module__}.{cls.__qualname__}"
-            result = db_count(errors=errors,
-                              table=sob_db_specs[name][0],
+            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+            tbl_name: str = sob_db_specs[cls_name][0]
+            if table_alias:
+                tbl_name = f"{tbl_name} {table_alias}"
+            if table_joins:
+                tbl_name += f" {table_joins}"
+            result = db_count(table=tbl_name,
+                              count_clause=count_clause,
                               where_clause=where_clause,
                               where_vals=where_vals,
                               where_data=where_data,
                               engine=db_engine,
                               connection=db_conn,
+                              errors=errors,
                               logger=logger)
         if errors and logger:
             logger.error(msg="; ".join(errors))
@@ -449,10 +506,16 @@ class PySob:
         return result
 
     @staticmethod
-    def exists(errors: list[str] = None,
+    def exists(table_alias: str = None,
+               table_joins: str = None,
+               where_clause: str = None,
+               where_vals: tuple = None,
                where_data: dict[str, Any] = None,
+               min_count: int = None,
+               max_count: int = None,
                db_engine: DbEngine = None,
                db_conn: Any = None,
+               errors: list[str] = None,
                logger: Logger = None) -> int | None:
 
         # inicialize the return variable
@@ -466,12 +529,21 @@ class PySob:
         cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
                                                     logger=logger)
         if not errors:
-            name: str = f"{cls.__module__}.{cls.__qualname__}"
-            result = db_exists(errors=errors,
-                               table=sob_db_specs[name][0],
+            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+            tbl_name: str = sob_db_specs[cls_name][0]
+            if table_alias:
+                tbl_name = f"{tbl_name} {table_alias}"
+            if table_joins:
+                tbl_name += f" {table_joins}"
+            result = db_exists(table=tbl_name,
+                               where_clause=where_clause,
+                               where_vals=where_vals,
                                where_data=where_data,
+                               min_count=min_count,
+                               max_count=max_count,
                                engine=db_engine,
                                connection=db_conn,
+                               errors=errors,
                                logger=logger)
         if errors and logger:
             logger.error(msg="; ".join(errors))
@@ -479,8 +551,8 @@ class PySob:
         return result
 
     @staticmethod
-    def retrieve(errors: list[str] = None,
-                 load_references: bool = False,
+    def retrieve(__references: list[type] = None,
+                 /,
                  table_alias: str = None,
                  table_joins: str = None,
                  where_clause: str = None,
@@ -493,6 +565,8 @@ class PySob:
                  limit_count: int = None,
                  db_engine: DbEngine = None,
                  db_conn: Any = None,
+                 committable: bool = None,
+                 errors: list[str] = None,
                  logger: Logger = None) -> list[Sob] | None:
 
         # inicialize the return variable
@@ -506,15 +580,20 @@ class PySob:
         cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
                                                     logger=logger)
         if not errors:
-            name: str = f"{cls.__module__}.{cls.__qualname__}"
-            sel_stmt: str = (f"SELECT {sob_db_specs[name][1]} "
-                             f"FROM {sob_db_specs[name][0]}")
+            # build the attributes list
+            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+            tbl_name: str = sob_db_specs[cls_name][0]
+            attrs: list[str] = []
+            for attr in sob_col_names.get(cls_name):
+                if table_alias:
+                    attr = f"{table_alias}.{attr}"
+                attrs.append(attr)
             if table_alias:
-                sel_stmt = sel_stmt.replace("SELECT", f"SELECT {table_alias}.") + f" {table_alias}"
+                tbl_name = f"{tbl_name} {table_alias}"
             if table_joins:
-                sel_stmt += f" {table_joins}"
-            recs: list[tuple[int | str]] = db_select(errors=errors,
-                                                     sel_stmt=sel_stmt,
+                tbl_name += f" {table_joins}"
+            sel_stmt: str = f"SELECT DISTINCT {', '.join(attrs)} FROM {tbl_name}"
+            recs: list[tuple[int | str]] = db_select(sel_stmt=sel_stmt,
                                                      where_clause=where_clause,
                                                      where_vals=where_vals,
                                                      where_data=where_data,
@@ -525,21 +604,48 @@ class PySob:
                                                      limit_count=limit_count,
                                                      engine=db_engine,
                                                      connection=db_conn,
+                                                     committable=committable,
+                                                     errors=errors,
                                                      logger=logger)
             if not errors:
                 # build the objects list
                 objs: list[Sob] = []
                 for rec in recs:
-                    # constructor of 'cls', a subclass of 'PySob'
-                    objs.append(cls(rec[0],
-                                    errors=errors,
-                                    load_references=load_references,
-                                    db_engine=db_engine,
-                                    db_conn=db_conn,
-                                    logger=logger))
+                    data: dict[str, Any] = {}
+                    for inx, attr in enumerate(sob_col_names.get(cls_name)):
+                        data[attr] = rec[inx]
+                    sob: type[Sob] = cls()
+                    sob.set(data=data)
+                    sob._is_new = False
                     if errors:
                         break
+                    objs.append(sob)
 
+                if __references:
+                    if SOB_MAX_THREADS > 1:
+                        task_futures: list[Future] = []
+                        with ThreadPoolExecutor(max_workers=SOB_MAX_THREADS) as executor:
+                            for obj in objs:
+                                # HAZARD: must not use 'db_conn' concurrently
+                                future: Future = executor.submit(obj.load_references,
+                                                                 __references,
+                                                                 # db_conn=db_conn,
+                                                                 db_engine=db_engine,
+                                                                 errors=errors)
+                                task_futures.append(future)
+
+                        # wait for all task futures to complete, then shutdown down the executor
+                        futures.wait(fs=task_futures)
+                        executor.shutdown(wait=False)
+
+                    else:
+                        for obj in objs:
+                            obj.load_references(__references,
+                                                db_engine=db_engine,
+                                                db_conn=db_conn,
+                                                errors=errors)
+                            if errors:
+                                break
                 if not errors:
                     result = objs
 
@@ -549,10 +655,51 @@ class PySob:
         return result
 
     @staticmethod
-    def erase(errors: list[str] = None,
+    def include(insert_data: dict[str, Any] = None,
+                return_cols: dict[str, Any] = None,
+                db_engine: DbEngine = None,
+                db_conn: Any = None,
+                committable: bool = None,
+                errors: list[str] = None,
+                logger: Logger = None) -> tuple | int | None:
+
+        # initialize the return variable
+        result: tuple | int | None = None
+
+        # make sure to have an errors list
+        if not isinstance(errors, list):
+            errors = []
+
+        # obtain the invoking class
+        cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
+                                                    logger=logger)
+        # delete specified rows
+        if not errors:
+            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+            tbl_name: str = sob_db_specs[cls_name][0]
+            result = db_insert(insert_stmt=f"INSERT INTO {tbl_name}",
+                               insert_data=insert_data,
+                               return_cols=return_cols,
+                               engine=db_engine,
+                               connection=db_conn,
+                               committable=committable,
+                               errors=errors,
+                               logger=logger)
+        if errors and logger:
+            logger.error(msg="; ".join(errors))
+
+        return result
+
+    @staticmethod
+    def erase(where_clause: str = None,
+              where_vals: tuple = None,
               where_data: dict[str, Any] = None,
+              min_count: int = None,
+              max_count: int = None,
               db_engine: DbEngine = None,
               db_conn: Any = None,
+              committable: bool = None,
+              errors: list[str] = None,
               logger: Logger = None) -> int | None:
 
         # initialize the return variable
@@ -565,15 +712,21 @@ class PySob:
         # obtain the invoking class
         cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
                                                     logger=logger)
-        # delete specified tuples
+        # delete specified rows
         if not errors:
-            name: str = f"{cls.__module__}.{cls.__qualname__}"
-            result: int = db_delete(errors=errors,
-                                    delete_stmt=f"DELETE FROM {sob_db_specs[name][0]}",
-                                    where_data=where_data,
-                                    engine=db_engine,
-                                    connection=db_conn,
-                                    logger=logger)
+            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+            tbl_name: str = sob_db_specs[cls_name][0]
+            result = db_delete(delete_stmt=f"DELETE FROM {tbl_name}",
+                               where_clause=where_clause,
+                               where_vals=where_vals,
+                               where_data=where_data,
+                               min_count=min_count,
+                               max_count=max_count,
+                               engine=db_engine,
+                               connection=db_conn,
+                               committable=committable,
+                               errors=errors,
+                               logger=logger)
         if errors and logger:
             logger.error(msg="; ".join(errors))
 
